use anyhow::Result;
use std::{fs::{File, self}, path::PathBuf, io::{BufReader, Write, self}, collections::HashMap};
use clap::{Parser, Subcommand, Args};
use xdg;

#[derive(Debug, Parser)]
struct Cli {
    #[command(subcommand)]
    action: Action,
}

#[derive(Debug, Subcommand)]
enum Action {
    New(NewProjectArgs),
    Delete(ProjectNameArgs),
    Path(ProjectNameArgs),
    List,
}

#[derive(Debug, Args)]
struct NewProjectArgs {
    project_name: String,
    project_path: String,
}

#[derive(Debug, Args)]
struct ProjectNameArgs {
    project_name: String,
}

type ProjectData = HashMap<String, String>;

fn main() {
    let cli = Cli::parse(); 

    let project_data = match load_or_default_project_data() {
        Ok(data) => data,
        Err(error) => {
            println!("An error occured while reading the project data file. {}", error);
            return;
        }
    };

    process_command(cli, project_data);
}

fn load_or_default_project_data() -> Result<ProjectData> {
    let data_file_path = xdg::BaseDirectories::with_prefix("projects")?.place_data_file("projects.json")?;
    create_default_data_file_if_needed(&data_file_path);
    let data_file = File::open(data_file_path)?;
    let reader = BufReader::new(data_file);

    let project_data: ProjectData = serde_json::from_reader(reader)?;

    return Ok(project_data);
}

fn save_project_data(project_data: ProjectData) -> Result<()> {
    let data_file_path = xdg::BaseDirectories::with_prefix("projects")?.place_data_file("projects.json")?;
    let mut data_file = File::create(data_file_path)?;

    let data_string = serde_json::to_string(&project_data)?;
    write!(data_file, "{}", data_string).expect("Failed to write new data file!");

    return Ok(());
}

fn create_default_data_file_if_needed(data_file_path: &PathBuf) {
    if data_file_path.exists() { return; }

    fs::create_dir_all(data_file_path.parent().unwrap()).ok();

    let mut data_file = File::create(data_file_path).expect("Error creating data file!");
    write!(data_file, "{{}}").expect("Error writing default content to data file!");
}

fn process_command(cli: Cli, project_data: ProjectData) -> Option<()> {
    match cli.action {
        Action::New(NewProjectArgs { project_name, project_path }) => new_project(project_name, project_path, project_data),
        Action::Delete(ProjectNameArgs {project_name}) => delete_project(project_name, project_data),
        Action::Path(ProjectNameArgs {project_name}) => project_path(project_name, project_data),
        Action::List => project_list(project_data),
    };

    return None;
}

fn new_project(project_name: String, project_path: String, mut project_data: ProjectData) {
    if project_data.contains_key(&project_name) {
        println!("Project with the name {} already exists!", project_name);
        return;
    }

    let project_path = std::path::Path::new(&project_path);

    if !project_path.exists() {
        fs::create_dir_all(&project_path).ok();
    }

    let project_path = match project_path.canonicalize() {
        Ok(path) => path,
        Err(error) => {
            println!("An error occured while resolving the given path. {}", error);
            return;
        }
    }; 

    project_data.insert(project_name.to_owned(), project_path.into_os_string().into_string().expect("Invalid path!"));

    save_project_data(project_data).expect("Error while saving data file!");
    println!("Successfully created new project {}!", project_name);
}

fn delete_project(project_name: String, mut project_data: ProjectData) {
    if !project_data.contains_key(&project_name) {
        println!("No project with the name {} exists!", project_name);
        return;
    }

    let project_path_string = &project_data[&project_name];
    let project_path = std::path::Path::new(project_path_string);

    println!("Are you sure you want to remove project with name {}? WARNING: This will also delete all files at {}! (y/N) ", project_name, project_path_string);
    let mut buffer = String::new();
    io::stdin().read_line(&mut buffer).expect("Error while reading stdin!");
    if buffer.trim() != "y" {
        println!("Aborting...");
        return;
    }

    fs::remove_dir_all(project_path).expect("Error while deleting project directory!");
    
    project_data.remove(&project_name);

    save_project_data(project_data).expect("Error while saving data file!");
    println!("Successfully removed project {}!", project_name);
}

fn project_path(project_name: String, project_data: ProjectData) {
    let project_path = project_data.get(&project_name);

    if let Some(project_path) = project_path {
        println!("{}", project_path);
    } else {
        println!("No such project!")
    };
}

fn project_list(project_data: ProjectData) {
    for project_name in project_data.keys() {
        println!("{}", project_name);
    }
}

